import time
import sys
import datetime

from ..game.team import Team

class BasicViewer:
    def __init__(self, board_size, board_states, colors=True):
        self.board_size = board_size
        self.board_states = board_states
        self.colors = colors

    def play(self, delay=0.5, keep_history=False):
        print('')

        for state_index in range(len(self.board_states)):
            self.view(state_index)
            time.sleep(delay)
            if not keep_history:
                self.clear()

        self.view(-1)

    def play_synchronized(self, poison_pill, delay=0.5, keep_history=False):
        print('')
        
        state_index = 0
        last_time = datetime.datetime.now().timestamp()
        while state_index < len(self.board_states) or not poison_pill.is_set():
            while len(self.board_states) <= state_index or datetime.datetime.now().timestamp() - last_time < delay:
                time.sleep(0.1)
            if not keep_history and state_index > 0:
                self.clear()
            self.view(state_index)
            last_time = datetime.datetime.now().timestamp()
            state_index += 1
            
            



    def clear(self):
        for i in range(self.board_size + 1):
            sys.stdout.write("\033[F")  # back to previous line
            sys.stdout.write("\033[K")  # clear line
    
    def view(self, index=-1):
        print(self.view_board(self.board_states[index]))

    def view_board(self, board):
        new_board = ''
        for i in range(self.board_size):
            for j in range(self.board_size):
                if board[i][j]:
                    new_board += '['
                    if self.colors:
                        if board[i][j].team == Team.WHITE:
                            new_board += '\033[1m\u001b[37m'
                        else:
                            new_board += '\033[1m\u001b[36m'
                    new_board += str(board[i][j])
                    if self.colors:
                        new_board += '\033[0m\u001b[0m'
                    new_board += '] '
                else:
                    new_board += '[    ] '
            new_board += '\n'
        return new_board



class Runner:
    def __init__(self, code_container, methods):
        self.code = code_container.code
        self.globals = {}
        self.locals = {}
        for key, method in methods.items():
            self.globals[key] = method
        
        exec(self.code, self.globals, self.locals)
        self.globals.update(self.locals)

    def do_turn(self):
        if 'turn' in self.locals and isinstance(self.locals['turn'], type(lambda: 1)):
            exec(self.locals['turn'].__code__, self.globals, self.locals)
        else:
            raise Exception("File does not have a 'turn' method.")



from typing import List, Optional, Tuple, Union

from battlehack20.engine.game.game import Team, RobotType

# The stubs in this file make it possible for editors to auto-complete the global methods
# They can be imported using "from battlehack20.stubs import *"
# This import is preprocessed away before instrumenting the code

# The dummy implementations in this file exist so that editors won't give warnings like
# "Assigning result of a function call, where the function has no return"


def log(msg: str) -> None:
    """
    Type-agnostic method.
    Logs a message.
    """
    log(msg)


def get_board_size() -> int:
    """
    Type-agnostic method.
    Returns the board size.
    """
    return get_board_size()


def get_bytecode() -> int:
    """
    Type-agnostic method.
    Returns the number of bytecodes left.
    """
    return get_bytecode()


def get_team() -> Team:
    """
    Type-agnostic method.
    Returns the robot’s team, either `Team.WHITE` or `Team.BLACK`.
    """
    return get_team()


def get_type() -> RobotType:
    """
    Type-agnostic method.
    Returns the robot’s type, either `RobotType.OVERLORD` or `RobotType.PAWN`.
    """
    return get_type()


def check_space(row: int, col: int) -> Union[Team, bool]:
    """
    Type-agnostic method.
    Returns `False` if there is no robot at the location, the team of the robot if there is one there,
    and throws a `RobotError` if outside the vision range.
    """
    return check_space(row, col)


def get_board() -> List[List[Optional[Team]]]:
    """
    Overlord method.
    Returns the current state of the board as an array of `Team.WHITE`, `Team.BLACK`, and `None`, representing
    white-occupied, black-occupied, and empty squares, respectively.
    """
    return get_board()


def spawn(row: int, col: int) -> None:
    """
    Overlord method.
    Spawns a pawn at the given location, but throws a `RobotError` if the pawn is not spawned at the edge on your
    side of the board, or if you have already spawned a pawn in this turn.
    """
    spawn(row, col)


def capture(row: int, col: int) -> None:
    """
    Pawn method.
    Captures an enemy piece at the given location, but throws a `RobotError` if the there is not an enemy pawn there
    or if the location is not diagonally in front of you.
    """
    capture(row, col)


def get_location() -> Tuple[int, int]:
    """
    Pawn method.
    Returns a `(row, col)` tuple of the robot’s location.
    """
    return get_location()


def move_forward() -> None:
    """
    Pawn method.
    Moves forward one step, but throws a `RobotError` if you have already moved, if the location is outside the board
    or if there is another pawn in front of you.
    """
    move_forward()


def sense() -> List[Tuple[int, int, Team]]:
    """
    Pawn method.
    Returns a list of tuples of the form `(row, col, robot.team)` visible to this robot (excluding yourself),
    that is, if `max(|robot.x - other.x|, |robot.y - other.y|) <= 2`.
    """
    return sense()


class GameConstants:

    # the maximum number of rounds, until the winner is decided by a coinflip
    MAX_ROUNDS = 500

    # the board size
    BOARD_SIZE = 16

    # the default seed
    DEFAULT_SEED = 1337



import os
import re

class CodeContainer:
    def __init__(self, code):
        self.code = code
        self.code = compile(code, "<string>", "exec")
                
    @classmethod
    def from_directory(cls, dirname):
        file = os.path.join(dirname, "bot.py")
        with open(file) as f:
            code = f.read()
        code = cls.preprocess(code)
        return cls(code)

    @classmethod
    def preprocess(cls, content):
        """
        Strips battlehack20.stubs imports from the code.
        It removes lines containing one of the following imports:
        - from battlehack20.stubs import *
        - from battlehack20.stubs import a, b, c
        The regular expression that is used also supports non-standard whitespace styles like the following:
        - from battlehack20.stubs import a,b,c
        - from  battlehack20.stubs  import  a,  b,  c
        Go to https://regex101.com/r/bhAqFE/6 to test the regular expression with custom input.
        """

        pattern = r'^([ \t]*)from([ \t]+)battlehack20\.stubs([ \t]+)import([ \t]+)(\*|([a-zA-Z_]+([ \t]*),([ \t]*))*[a-zA-Z_]+)([ \t]*)$'

        # Replace all stub imports
        while True:
            match = re.search(pattern, content, re.MULTILINE)

            if match is None:
                break

            # Remove the match from the content
            start = match.start()
            end = match.end()
            content = content[0:start] + content[end:]

        return content

    @classmethod
    def preprocess(cls, content):
        """
        Strips battlehack20.stubs imports from the code.
        It removes lines containing one of the following imports:
        - from battlehack20.stubs import *
        - from battlehack20.stubs import a, b, c
        The regular expression that is used also supports non-standard whitespace styles like the following:
        - from battlehack20.stubs import a,b,c
        - from  battlehack20.stubs  import  a,  b,  c
        Go to https://regex101.com/r/bhAqFE/6 to test the regular expression with custom input.
        """
        # Code copied directly from battlehack20 engine

        pattern = r'^([ \t]*)from([ \t]+)battlehack20\.stubs([ \t]+)import([ \t]+)(\*|([a-zA-Z_]+([ \t]*),([ \t]*))*[a-zA-Z_]+)([ \t]*)$'

        # Replace all stub imports
        while True:
            match = re.search(pattern, content, re.MULTILINE)

            if match is None:
                break

            # Remove the match from the content
            start = match.start()
            end = match.end()
            content = content[0:start] + content[end:]

        return content



from setuptools import setup, find_packages
from collections import OrderedDict

long_description="""
This is a minimal engine for the Battlehack20 game.
It lacks the secure of the original engine, but makes up for it
by running 30 times faster.
Read more at the Battlehack website: https://bh2020.battlecode.org.
"""

setup(name='battlehack20-minimal',
      version="1.0.5",
      description='Battlehack 2020 fancy viewer.',
      author='cooljoseph',
      long_description=long_description,
      author_email='camacho.joseph@gmail.com',
      url="https://bh2020.battlecode.org",
      license='GNU General Public License v3.0',
      packages=find_packages(),
      project_urls=OrderedDict((
          ('Code', 'https://github.com/cooljoseph1/battlehack20-minimal'),
          ('Documentation', 'https://github.com/cooljoseph1/battlehack20-minimal')
      )),
      install_requires=[],
      python_requires='>=3, <3.8',
      zip_safe=False,
      include_package_data=True
)



import random
import traceback
from .constants import GameConstants
from ..container.runner import Runner
from .robot import Robot
from .robottype import RobotType
from .team import Team

class Game:
    def __init__(self, code, board_size=GameConstants.BOARD_SIZE, max_rounds=GameConstants.MAX_ROUNDS, 
                 seed=GameConstants.DEFAULT_SEED, sensor_radius=2, debug=False):
        
        random.seed(seed)

        self.code = code

        self.debug = debug
        self.running = True
        self.winner = None

        self.robot_count = 0
        self.queue = {}
        self.leaders = []

        self.sensor_radius = sensor_radius
        self.board_size = board_size
        self.board = [[None] * self.board_size for _ in range(self.board_size)]
        self.round = 0
        self.max_rounds = max_rounds

        self.robot = None # Robot that is currently doing a turn
        self.shared_methods = {"Team": Team,
                            "RobotType": RobotType,
                            "log": self.log,
                            "get_board_size": self.get_board_size,
                            "get_bytecode": self.get_bytecode,
                            "get_team": self.get_team,
                            "get_type": self.get_type,
                            "check_space": self.check_space}
        self.overlord_methods = {"get_board": self.get_board,
                                 "spawn": self.spawn}
        self.pawn_methods = {"capture": self.capture,
                             "get_location": self.get_location,
                             "move_forward": self.move_forward,
                             "sense": self.sense}
        self.overlord_methods.update(self.shared_methods)
        self.pawn_methods.update(self.shared_methods)

        self.new_robot(None, None, Team.WHITE, RobotType.OVERLORD)
        self.new_robot(None, None, Team.BLACK, RobotType.OVERLORD)

        
        self.board_states = []

    def turn(self):
        self.round += 1

        if self.round > self.max_rounds:
            self.check_over()

        for i in range(self.robot_count):
            if i in self.queue:
                self.robot = self.queue[i]

                try:
                    self.robot.runner.do_turn()
                except Exception as e:
                    if self.debug:
                        traceback.print_exc()
                self.check_over()

        if self.running:
            self.queue[0], self.queue[1] = self.queue[1], self.queue[0] # Alternate spawn order of Overlords
            self.board_states.append([row[:] for row in self.board])

    def new_robot(self, row, col, team, robot_type):
        methods = self.overlord_methods if robot_type == RobotType.OVERLORD else self.pawn_methods
        code = self.code[0] if team == Team.WHITE else self.code[1]
        runner = Runner(code, methods)
        
        robot = Robot(row, col, team, self.robot_count, robot_type, runner)
        
        self.queue[robot.id] = robot
        if robot.type != RobotType.OVERLORD:
            self.board[robot.row][robot.col] = robot
        
        self.robot_count += 1

    def delete_robot(self, i):
        robot = self.queue[i]
        if robot.type != RobotType.OVERLORD:
            self.board[robot.row][robot.col] = None
        del self.queue[i]

    def is_on_board(self, row, col):
        return 0 <= row < self.board_size and 0 <= col < self.board_size

    def check_over(self):
        white, black = 0, 0
        for col in range(self.board_size):
            if self.board[0][col] and self.board[0][col].team == Team.BLACK: black += 1
            if self.board[self.board_size - 1][col] and self.board[self.board_size - 1][col].team == Team.WHITE: white += 1

        # End the game in the case of three turns where nothing changes
        #if self.round >= 4 and self.board_states[-1] == self.board_states[-2] == self.board_states[-3]:
        #    self.round = self.max_rounds + 1
        
        if self.round > self.max_rounds:
            self.running = False
            i = 0
            while white == black and i < self.board_size - 1:
                i += 1
                white, black = 0, 0
                for col in range(self.board_size):
                    if self.board[i][col] and self.board[i][col].team == Team.BLACK: black += 1
                    if self.board[self.board_size - 1 - i][col] and self.board[self.board_size - 1 - i][col].team == Team.WHITE: white += 1
            if white > black:
                self.winner = Team.WHITE
            elif white < black:
                self.winner = Team.BLACK
            else:
                self.winner = random.choice([Team.WHITE, Team.BLACK])

        if white >= (self.board_size + 1) // 2:
            self.running = False
            self.winner = Team.WHITE

        if black >= (self.board_size + 1) // 2:
            self.running = False
            self.winner = Team.BLACK

        if not self.running:
            self.board_states.append([row[:] for row in self.board])
            self.process_over()

    def process_over(self):
        """
        Helper method to process once a game is finished (e.g. deleting robots)
        """
        for i in range(self.robot_count):
            if i in self.queue:
                self.delete_robot(i)

    ### GENERAL METHODS ###
    def log(self, *args, **kwargs):
        if self.debug:
            print(*args, **kwargs)
    
    def get_board_size(self):
        return self.board_size

    def get_bytecode(self):
        return 1000000

    def get_team(self):
        return self.robot.team

    def get_type(self):
        return self.robot.type
    
    def check_space(self, row, col):
        return self.board[row][col].team if self.board[row][col] else False

    ### OVERLORD METHODS ###
    def get_board(self):
        return [[robot.team if robot else None for robot in row] for row in self.board]
    
    def spawn(self, row, col):
        self.new_robot(row, col, self.robot.team, RobotType.PAWN)
    
    ### PAWN METHODS ###
    
    def capture(self, new_row, new_col):
        self.delete_robot(self.board[new_row][new_col].id)
        self.board[new_row][new_col] = self.robot
        self.board[self.robot.row][self.robot.col] = None
        self.robot.row = new_row
        self.robot.col = new_col

    def get_location(self):
        return self.robot.row, self.robot.col

    def move_forward(self):
        if self.robot.team == Team.WHITE:
            new_row, new_col = self.robot.row + 1, self.robot.col
        else:
            new_row, new_col = self.robot.row - 1, self.robot.col
            
        self.board[new_row][new_col] = self.robot
        self.board[self.robot.row][self.robot.col] = None
        self.robot.row = new_row
        self.robot.col = new_col

    def sense(self):
        robots = []
        for new_row in range(self.robot.row - self.sensor_radius, self.robot.row + self.sensor_radius + 1):
            for new_col in range(self.robot.col - self.sensor_radius, self.robot.col + self.sensor_radius + 1):
                if new_row == self.robot.row and new_col == self.robot.col:
                    continue
                if not self.is_on_board(new_row, new_col):
                    continue
                if self.board[new_row][new_col]:
                    robots.append((new_row, new_col, self.board[new_row][new_col].team))
        return robots



from .robottype import RobotType

class Robot:
    def __init__(self, row, col, team, id, type, runner):
        self.row = row
        self.col = col
        self.team = team
        self.id = id
        self.type = type
        self.runner = runner
